# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
"""
anycast healthchecker daemon
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A healthchecker for Anycasted services.

"""
__title__ = 'anycast_healthchecker'
__author__ = 'Pavlos Parissis'
__license__ = 'Apache 2.0'
__version__ = '0.5.10'
__copyright__ = 'Copyright 2015-2016 Pavlos Parissis'
